IMPLEMENTATION MODULE ctype;
__IMP_SWITCHES__
#ifdef HM2
#ifdef __LONG_WHOLE__
(*$!i+: Modul muss mit $i- uebersetzt werden! *)
(*$!w+: Modul muss mit $w- uebersetzt werden! *)
#else
(*$!i-: Modul muss mit $i+ uebersetzt werden! *)
(*$!w-: Modul muss mit $w+ uebersetzt werden! *)
#endif
#endif
(*****************************************************************************)
(* 20-Aug-93, Holger Kleinschmidt                                            *)
(*****************************************************************************)

VAL_INTRINSIC

TYPE
  CType  = (LOWER, UPPER, DIGIT, XDIGIT, PUNCT, CONTROL, SPACE);

  CClass = PACKEDSET OF CType;

#if only_subrange_index
  CClasses = ARRAY [0C..377C] OF CClass;
#else
  CClasses = ARRAY CHAR OF CClass;
#endif

CONST
  NOALPHA = CClass{DIGIT..SPACE};
  NOALNUM = CClass{PUNCT..SPACE};
  GRAPH   = CClass{LOWER..PUNCT};

VAR
  ctype : CClasses;

(*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*)

PROCEDURE isalnum ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(NOT(ctype[c] <= NOALNUM));
END isalnum;

(*---------------------------------------------------------------------------*)

PROCEDURE isalpha ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(NOT(ctype[c] <= NOALPHA));
END isalpha;

(*---------------------------------------------------------------------------*)

PROCEDURE isascii ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(c <= 177C);
END isascii;

(*---------------------------------------------------------------------------*)

PROCEDURE iscntrl ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(CONTROL IN ctype[c]);
END iscntrl;

(*---------------------------------------------------------------------------*)

PROCEDURE isdigit ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(DIGIT IN ctype[c]);
END isdigit;

(*---------------------------------------------------------------------------*)

PROCEDURE isgraph ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(ctype[c] <= GRAPH);
END isgraph;

(*---------------------------------------------------------------------------*)

PROCEDURE islower ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(LOWER IN ctype[c]);
END islower;

(*---------------------------------------------------------------------------*)

PROCEDURE isprint ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(NOT(CONTROL IN ctype[c]));
END isprint;

(*---------------------------------------------------------------------------*)

PROCEDURE ispunct ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(PUNCT IN ctype[c]);
END ispunct;

(*---------------------------------------------------------------------------*)

PROCEDURE isspace ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(SPACE IN ctype[c]);
END isspace;

(*---------------------------------------------------------------------------*)

PROCEDURE isupper ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(UPPER IN ctype[c]);
END isupper;

(*---------------------------------------------------------------------------*)

PROCEDURE isxdigit ((* EIN/ -- *) c : CHAR ): BOOLEAN;
(*T*)
BEGIN
 RETURN(XDIGIT IN ctype[c]);
END isxdigit;

(*---------------------------------------------------------------------------*)

PROCEDURE tolower ((* EIN/ -- *) c : CHAR ): CHAR;
(**)
BEGIN
 IF UPPER IN ctype[c] THEN
   RETURN(CHR(ORD(c) + 32));
 ELSE
   RETURN(c);
 END;
END tolower;

(*---------------------------------------------------------------------------*)

PROCEDURE toupper ((* EIN/ -- *) c : CHAR ): CHAR;
(**)
BEGIN
 IF LOWER IN ctype[c] THEN
   RETURN(CHR(ORD(c) - 32));
 ELSE
   RETURN(c);
 END;
END toupper;

(*---------------------------------------------------------------------------*)

PROCEDURE tocard ((* EIN/ -- *) c : CHAR ): CARDINAL;
(**)
BEGIN
 IF DIGIT IN ctype[c] THEN
   RETURN(VAL(CARDINAL,ORD(c) - ORD('0')));
 ELSIF UPPER IN ctype[c] THEN
   RETURN(VAL(CARDINAL,ORD(c) - ORD('A') + 10));
 ELSIF LOWER IN ctype[c] THEN
   RETURN(VAL(CARDINAL,ORD(c) - ORD('a') + 10));
 ELSE
   RETURN(VAL(CARDINAL,ORD('Z') + 1)); (* ``ungueltiger'' Wert *)
 END;
END tocard;

(*---------------------------------------------------------------------------*)

PROCEDURE todigit ((* EIN/ -- *) i : CARDINAL ): CHAR;
(*T*)
BEGIN
 IF i <= 9 THEN
   RETURN(CHR(i + VAL(CARDINAL,ORD('0'))));
 ELSE
   RETURN(CHR(i + VAL(CARDINAL,ORD('A')) - 10));
 END;
END todigit;

(*===========================================================================*)
#if !ISO_value_constructor
VAR c : CHAR;
#endif

BEGIN (* ctype *)
#if ISO_value_constructor
 ctype := CClasses{CClass{CONTROL}       BY 9,
                   CClass{CONTROL,SPACE} BY 5,
                   CClass{CONTROL}       BY 18,
                   CClass{SPACE}         BY 1,
                   CClass{PUNCT}         BY 15,
                   CClass{DIGIT,XDIGIT}  BY 10,
                   CClass{PUNCT}         BY 7,
                   CClass{XDIGIT,UPPER}  BY 6,
                   CClass{UPPER}         BY 20,
                   CClass{PUNCT}         BY 6,
                   CClass{XDIGIT,LOWER}  BY 6,
                   CClass{LOWER}         BY 20,
                   CClass{PUNCT}         BY 4,
                   CClass{CONTROL}       BY 1,
                   CClass{PUNCT}         BY 128
                  };
#else
 FOR c := 0C TO 10C DO
   ctype[c] := CClass{CONTROL};
 END;
 FOR c := 11C TO 15C DO
   ctype[c] := CClass{CONTROL, SPACE};
 END;
 FOR c := 16C TO 37C DO
   ctype[c] := CClass{CONTROL};
 END;
 ctype[' '] := CClass{SPACE};
 FOR c := 41C TO 57C DO
   ctype[c] := CClass{PUNCT};
 END;
 FOR c := '0' TO '9' DO
   ctype[c] := CClass{DIGIT, XDIGIT};
 END;
 FOR c := 72C TO 100C DO
   ctype[c] := CClass{PUNCT};
 END;
 FOR c := 'A' TO 'F' DO
   ctype[c] := CClass{XDIGIT, UPPER};
 END;
 FOR c := 'G' TO 'Z' DO
   ctype[c] := CClass{UPPER};
 END;
 FOR c := 133C TO 140C DO
   ctype[c] := CClass{PUNCT};
 END;
 FOR c := 'a' TO 'f' DO
   ctype[c] := CClass{XDIGIT, LOWER};
 END;
 FOR c := 'g' TO 'z' DO
   ctype[c] := CClass{LOWER};
 END;
 FOR c := 173C TO 176C DO
   ctype[c] := CClass{PUNCT};
 END;
 ctype[177C] := CClass{CONTROL};
 FOR c := 200C TO 377C DO
   ctype[c] := CClass{};
 END;
#endif
END ctype.
